import inspect
import functools
from .dot import dotgraph


class Path(list):
    '''Store :class:`Edge` objects connecting two :class:`Node`
    objects.

    Parameters
    ----------
    edges : iterable, optional
        All :class:`Edge` objects in this :class:`Path`. If left
        out, the path is initialized empty.

    Attributes
    ----------
    edges
    '''

    def __str__(self):
        return dotgraph(edges=self)

    def __add__(self, other):
        return Path(super().__add__(other))

    def __repr__(self):
        return object.__repr__(self)

    def __lt__(self, other):
        if isinstance(other, Path):
            return self.weight < other.weight
        return NotImplemented

    @property
    def weight(self):
        '''Sum of the weights of all of the :class:`Edge`
        objects in this :class:`Path`.

        :type: numerical
        '''
        return sum(edge.weight for edge in self)


class _IncompleteSearch(Exception):
    pass


def memoize(shortest_path_func):
    '''Cache a path-finding function that expects two input parameters.

    The path-finding function may additionally accept some number of
    private parameters that don't affect cache.

    Examples
    --------
    Accessing cache.

    >>> shortest_path.cache  # see note for supported functions
    {...}


    Clearing cache.

    >>> shortest_path.cache_clear()  # see note for supported functions
    >>> shortest_path.cache
    {}

    Note
    ----
    Supported path-finding functions:

    - :meth:`shortest_path`
    - :meth:`shortest_path_through_network`
    - :meth:`path_exists`
    '''
    memo = {}
    all_params = iter(inspect.signature(shortest_path_func).parameters)
    param1_name = next(all_params)
    param2_name = next(all_params)

    @functools.wraps(shortest_path_func)
    def memoized_shortest_path_func(*args, **kwargs):
        try:
            param1 = args[0]
        except IndexError:  # both cacheable params passed as kwargs
            param1 = kwargs[param1_name]
            param2 = kwargs[param2_name]
        else:
            try:
                param2 = args[1]
            except IndexError:  # second cacheable param passed as kwarg
                param2 = kwargs[param2_name]

        cachekey = tuple(frozenset(param) if isinstance(param, set)
                         else param for param in [param1, param2])

        try:
            return memo[cachekey]
        except KeyError:
            try:
                path = shortest_path_func(*args, **kwargs)
            except _IncompleteSearch:
                return None  # but don't cache
            else:
                memo[cachekey] = path
                return path

    memoized_shortest_path_func.cache_clear = memo.clear
    memoized_shortest_path_func.cache = memo
    # pylint: disable=no-member
    memoized_shortest_path_func.__doc__ += '''
    Note
    ----
    This function is cached by the :meth:`memoize` function.
    '''
    return memoized_shortest_path_func


@memoize
def shortest_path(start, end, _visited=None,
                  _tail_weight=None,  # weight from the start of the best path
                  _best_path_weight=None):
    '''Find the shortest path between ``start`` and ``end``.

    Parameters
    ----------
    start: :class:`Node`
    end: :class:`Node`

    Returns
    -------
    :class:`Path` ``None``
        If a path exists from ``start`` to ``end``, return that
        :class:`Path` object. Otherwise, return ``None``.
    '''
    if start is end:
        return Path()
    if _visited is None:
        _visited = set()
    edges_were_skipped = False

    for edge in start.edges:
        if edge.node2 not in _visited:
            try:
                new_tail_weight = edge.weight + _tail_weight
            except TypeError:  # catches _tail_weight is None
                # haven't found a best --> we MUST keep going
                new_tail_weight = None  # still no tail
            else:
                # move on if weight of a path down this edge will exceed best
                if new_tail_weight >= _best_path_weight:
                    edges_were_skipped = True
                    continue
            path = shortest_path(
                edge.node2, end,
                _visited=_visited | {start},
                _tail_weight=new_tail_weight,
                _best_path_weight=_best_path_weight,
            )
            try:
                best_path = Path([edge]) + path
            except TypeError:  # no path found
                pass
            else:
                _best_path_weight = best_path.weight
                # ensure future calls will keep track of tail:
                _tail_weight = 0  # change it from None to 0
    try:
        return best_path
    except NameError:  # never found a path
        if edges_were_skipped:
            # the exception gets handled by memoize() wrapper, returning None
            # but NOT caching result (a path may exist!)
            raise _IncompleteSearch('although a path may exist, some searches '
                                    'aborted because the weight exceeded the '
                                    'current best.')
        return None


@memoize
def shortest_path_through_network(start, network, _visited=None,
                                  # weight from the start of the best path
                                  _tail_weight=None,
                                  _best_path_weight=None):
    '''Find the shortest path from ``start`` through all other
    :class:`Node` objects in ``network``.

    Parameters
    ----------
    start : :class:`Node`
        Start of the returned :class:`Path`.
    network : :class:`Network`
        Fully-connected network through which the returned
        :class:`Path` travels.

    Returns
    -------
    :class:`Path` ``None``
        If a path exists from ``start`` to ``end``, return that
        :class:`Path` object. Otherwise, return ``None``.
    '''
    try:
        reduced_set = network.all_nodes - {start}
    except AttributeError:  # network is set of Node, not Network
        reduced_set = network - {start}
    if not reduced_set:
        return Path()
    if _visited is None:
        _visited = set()
    edges_were_skipped = False

    for edge in start.edges:
        if edge.node2 not in _visited:
            try:
                new_tail_weight = edge.weight + _tail_weight
            except TypeError:  # catches _tail_weight is None
                # haven't found a best --> we MUST keep going
                new_tail_weight = None  # still no tail
            else:
                # move on if weight of a path down this edge will exceed best
                if new_tail_weight >= _best_path_weight:
                    edges_were_skipped = True
                    continue
            path = shortest_path_through_network(
                edge.node2, reduced_set,
                _visited=_visited | {start},
                _tail_weight=new_tail_weight,
                _best_path_weight=_best_path_weight,
            )
            try:
                best_path = Path([edge]) + path
            except TypeError:  # no path found
                pass
            else:
                _best_path_weight = best_path.weight
                # ensure future calls will keep track of tail:
                _tail_weight = 0  # change it from None to 0

    try:
        return best_path
    except NameError:  # never found a path
        if edges_were_skipped:
            # the exception gets handled by memoize() wrapper, returning None
            # but NOT caching result (a path may exist!)
            raise _IncompleteSearch('although a path may exist, some searches '
                                    'aborted because the weight exceeded the '
                                    'current best.')
        return None


@memoize
def path_exists(start, end, _visited=None):
    '''Check if a path exists between ``start`` and ``end``.

    Parameters
    ----------
    start : :class:`Node`
    end : :class:`Node`

    Returns
    -------
    bool
        ``True`` if a path exists, otherwise ``False``.
    '''

    if start == end:
        return True
    if _visited is None:
        _visited = set()

    for edge in start.edges:
        if edge.node2 not in _visited:
            if path_exists(edge.node2, end, _visited=_visited | {start}):
                return True
    return False



import re


def escape_dot_id(string):
    '''Surround in double quotes and escape all double quotes.

    Parameters
    ----------
    string : str
        ID to escape.

    Returns
    -------
    str
        A valid, escaped ID for a DOT graph.

    Example
    -------
    >>> escape_dot_id('A"B')
    '"A\\"B"'
    '''

    return '"' + re.sub(r'([\\"])', r'\\\1', string) + '"'


def dotgraph(isolated_nodes=None, edges=None, name=''):
    '''Generate a DOT graph out of nodes and edges.

    Parameters
    ----------
    isolated_nodes : list of :class:`Node`, optional
        Isolated nodes to graph.
    edges : list of :class:`Edge`, optional
        Edges to graph.
    name : optional
        Name of the returned dotgraph.

    Returns
    -------
    str
        A valid DOT graph of all the given nodes and edges with
        name ``name``.

    Examples
    --------
    Graphing one isolated :class:`Node` object.

    >>> a = Node('A')
    >>> print(dotgraph(isolated_nodes=[a], name='My Graph'))
    graph "My Graph" {
        "A"
    }

    Graphing connected :class:`Node` objects.

    >>> b = Node('B')
    >>> c = Node('C')
    >>> b.connect(c, 5)
    >>> print(dotgraph(edges=b.edges))
    graph {
        "B" -- "C" [label=5]
    }
    '''
    if isolated_nodes is None:
        isolated_nodes = []
    if edges is None:
        edges = []

    nodes = '\n\t'.join([escape_dot_id(node.name) for node in isolated_nodes])
    middle = '\n\t' if isolated_nodes and edges else ''
    edges = '\n\t'.join([con.dot() for con in edges])
    return (f'graph {f"{escape_dot_id(name)} " if name else ""}'
            f'{{\n\t{nodes}{middle}{edges}\n}}')



import random
from .dot import dotgraph
from .dot import escape_dot_id
from .pathfinding import path_exists


class Node:
    '''Contain a named node, with weighted edges.

    Parameters
    ----------
    name
        name of this :class:`Node`.


    Example
    -------
    >>> a = Node('A')
    >>> b = Node('B')
    >>> a.connect(b, 3)
    >>> print(a)
    graph {
        "A" -- "B" [label=3]
    }

    Attributes
    ---------
    name
    edges : list of :class:`Edge`
        list of edges containing this :class:`Node`.
    '''

    def __init__(self, name):
        self.name = str(name)
        self.edges = []

    def __str__(self):
        if self.edges:
            return dotgraph(edges=self.edges)
        return dotgraph(isolated_nodes=[self])

    @property
    def degree(self):
        '''Deegree of this node (number of outgoing edges).

        :type: int
        '''
        return len(self.edges)

    def connect(self, other, weight=None):
        '''Add :class:`Edge` between ``self`` and ``other`` with
        ``weight``.

        Parameters
        ----------
        other : :class:`Node`
        weight : numerical, optional
        '''
        self.edges.append(Edge(self, other, weight))
        other.edges.append(Edge(other, self, weight))

    def disconnect(self, other, weight=None):
        '''Remove :class:`Edge` between ``self`` and ``other``
        with ``weight``.

        Parameters
        ----------
        other : :class:`Node`
        weight : numerical, optional
        '''
        self.edges.remove(Edge(self, other, weight))
        other.edges.remove(Edge(other, self, weight))

    def isolate(self):
        '''Disconnect from all connected :class:`Node` objects.'''
        self.edges = []


class Edge:
    '''Represent an optionally weighted edge between two
    :class:`Node` objects.

    Parameters
    ----------
    node1 : :class:`Node`
        first :class:`Node` in this :class:`Edge`.
    node2 : :class:`Node`
        second :class:`Node` in this :class:`Edge`.
    weight : numerical, optional
        weight of this :class:`Edge`.

    Attributes
    ----------
    node1
    node2
    weight
    '''

    def __init__(self, node1, node2, weight=None):
        self.node1 = node1
        self.node2 = node2
        self.weight = weight

    def __repr__(self):
        return (f'{self.__class__.__name__}({self.node1!r}, {self.node2!r}'
                f'{f", {self.weight!r}" if self.weight else ""})')

    def __str__(self):
        return dotgraph(edges=[self])

    def __hash__(self):
        return hash(self.node1) ^ hash(self.node2) ^ hash(self.weight)

    def __eq__(self, other):
        return (self.node1, self.node2, self.weight) == (other.node1,
                                                         other.node2,
                                                         other.weight)

    def reverse(self):
        '''
        Returns
        -------
        :class:`Edge`
            A :class:`Edge` with the same ``node1``, ``node2``,
            and ``weight`` but with ``node1`` and ``node2`` swapped.
        '''
        return Edge(self.node2, self.node1, self.weight)

    def dot(self):
        '''
        Returns
        -------
        str
            A DOT language representation of this :class:`Edge`
            object.
        '''
        unlabeled = (f'{escape_dot_id(self.node1.name)} -- '
                     f'{escape_dot_id(self.node2.name)}')
        if self.weight:
            return unlabeled + f' [label={self.weight}]'
        return unlabeled


class Network:
    '''Contain a network of interconnected :class:`Node` objects.

    Parameters
    ----------
    all_nodes : iterable, optional
        All nodes in this network. If left out, the network is
        initialized empty.
    name: optional
        The name of this network.

    Attributes
    ----------
    all_nodes
    name
    edges : list
        All edges in this :class:`Network`.
    isolated_nodes : set
        All nodes with no edges.
    '''

    def __init__(self, all_nodes=None, name=None):
        self.all_nodes = set(all_nodes) if all_nodes is not None else set()
        self.name = str(name) if name is not None else ''
        self.update()  # set self.isolated_nodes and self.edges

    def __str__(self):
        return dotgraph(self.isolated_nodes, self.edges, self.name)

    def __iter__(self):
        yield from self.all_nodes

    @property
    def strongly_connected(self):
        '''``True`` if every node in this network has a path to every
        other node.

        :type bool:
        '''
        node_a = random.sample(self, 1)[0]
        others = self - {node_a}
        for node in others:
            if not path_exists(node_a, node):
                return False
        return True

    def update(self):
        '''Update ``edges`` and ``isolated_nodes``, to be used
        when :class:`Node` objects in this network have changed.
        '''
        self.isolated_nodes = set()
        self.edges = []
        seen = set()  # store reverses of seen edges
        for node in self.all_nodes:
            if node.edges:
                for con in node.edges:
                    if con.reverse() in seen:
                        # just node2 of an already-stored edge --> ignore
                        continue
                    self.edges.append(con)
                    seen.add(con)
            else:
                self.isolated_nodes.add(node)


def generate_network(n_nodes=10, lower_bound=1, upper_bound=11,
                     edge_prob=0.8, strongly_connected=True):
    '''Create a :class:`Network` of  :class:`Node` objects.

    Parameters
    ----------
    n_nodes : int, optional
        Number of nodes in the returned network.
    lower_bound : int, optional
        Lower bound (inclusive) of range of edges' weights.
    upper_bound : int, optional
        Upper bound (exclusive) for range of edges' weights.
    edges_prob : float, optional
        Probability betweeen 0 and 1 of any two nodes being connected.
        If ``strongly_connected`` is set to ``True``, ``edges_prob``
        may be overridden to ensure a strongly connected network.
    strongly_connected : bool
        If ``False``, output does not need to be a strongly connected
        network.

    Returns
    -------
    :class:`Network`
        A network ``n_nodes`` interconnected :class:`Node` objects.
    '''

    nodes = {Node(f'Node {i}') for i in range(int(n_nodes))}
    done = set()

    for cur_node in nodes:
        done.add(cur_node)
        for other_node in nodes:
            if other_node not in done and random.random() < edge_prob:
                cur_node.connect(other_node, random.randint(
                    lower_bound, upper_bound - 1))
    if strongly_connected:  # all nodes in our network must reach all others
        node_a = random.sample(done, 1)[0]
        others = done - {node_a}
        for node in others:
            if not path_exists(node_a, node):
                node.connect(node_a, random.randint(
                    lower_bound, upper_bound - 1))
                path_exists.cache_clear()

    return Network(done)



'''pynetworks
==========

A Python package that provides structure for networks of interconnected
nodes using the DOT language for representation.
'''

from .networks import Edge
from .networks import Network
from .networks import Node
from .networks import generate_network
from .dot import dotgraph
from .dot import escape_dot_id
from .pathfinding import Path
from .pathfinding import memoize
from .pathfinding import path_exists
from .pathfinding import shortest_path
from .pathfinding import shortest_path_through_network


__version__ = "0.6.0"

__all__ = ['Edge',
           'Network',
           'Node',
           'generate_network',
           'dotgraph',
           'escape_dot_id',
           'Path',
           'memoize',
           'path_exists',
           'shortest_path',
           'shortest_path_through_network'
           ]
